"""Path Condition"""
import copy
from debugutils import debugprint
from NodeType import NodeType
from VisitedNode import VisitedNode


class PathCondition:
	"""Class representing a path condition."""
	def __init__(self, left_node, right_node, relation_label, node_type, parameter_variable,  start_variable = None, start_constant= None, dest_variable = None, dest_constant = None):
		"""constructor"""

		#========== allowed argument types and interactions ==========
		# left node (non-leaf)
		if type(left_node) != type(self) and left_node != None:
			raise Exception("Argument 'left_node' must be either a PathCondition or None")
		# right node (non-leaf concatenation)
		if type(right_node) != type(self) and right_node != None:
			raise Exception("Argument 'right_node' must be either a PathCondition or None")
		# relationship label (leaf)
		if type(relation_label) != type("") and relation_label != None:
			raise Exception("Argument 'relation_label' must be either a String or None")
		# node type
		if node_type == NodeType.relation:
			# NodeType.relation indicates a leaf which requires:
			#	- left_node to be None
			#	- right_node to be None
			#	- relation_label to be a non-empty string
			if left_node != None:
				raise Exception("Argument 'left_node' must be None if argument 'node_type' is NodeType.relation")
			elif right_node != None:
				raise Exception("Argument 'right_node' must be None if argument 'node_type' is NodeType.relation")
			elif relation_label == None:
				raise Exception("Argument 'relation_label' must not be None if argument 'node_type' is NodeType.relation")
			elif relation_label == "":
				raise Exception("Argument 'relation_label' must not be "" if argument 'node_type' is NodeType.relation")
		elif node_type == NodeType.concatenate:
			# NodeType.concatenate indicates a binary operator which requires:
			#	- left_node to be a PathCondition
			#	- right_node to be a PathCondition
			#	- relation_label to be None
			if left_node == None:
				raise Exception("Argument 'left_node' must be a PathCondition if argument 'node_type' is NodeType.concatenate")
			elif right_node == None:
				raise Exception("Argument 'right_node' must be a PathCondition if argument 'node_type' is NodeType.concatenate")
			elif relation_label != None:
				raise Exception("Argument 'relation_label' must be None if argument 'node_type' is NodeType.concatenate")
		elif node_type == NodeType.reverse or node_type == NodeType.plus or node_type == NodeType.star:
			# NodeType.reverse, NodeType.plus and NodeType.star indicate unary operators requiring:
			#	- left_node to be a PathCondition
			#	- right_node to be None
			#	- relation_label to be None
			if left_node == None:
				raise Exception("Argument 'left_node' must be a PathCondition for this 'node_type'")
			elif right_node != None:
				raise Exception("Argument 'right_node' must be None for this 'node_type'")
			elif relation_label != None:
				raise Exception("Argument 'relation_label' must be None for this 'node_type'")
		else:
			raise Exception("Argument 'node_type' must specify a node_type value")

		#========== set internal data ==========
		self.left_node = left_node
		self.right_node = right_node
		self.relation_label = relation_label
		self.node_type = node_type
		self.parameter_variable = parameter_variable
		self.start_variable = start_variable
		self.start_constant = start_constant
		self.dest_variable = dest_variable
		self.dest_constant = dest_constant

	def to_string(self):
		"""Returns the path condition string"""
		_latex_format = False # used for modifying the output to _latex_format formatted strings (command line strings if False)

		# if path condition contains a path condition in the first slot
		if type(self.left_node) == type(self):
			if self.node_type == NodeType.relation:
				return self.relation_label.to_string()
			elif self.node_type == NodeType.reverse:
				if _latex_format:
					return "\overline{{{0}}}".format(self.left_node.to_string())
				else:
					return "¬{0}".format(self.left_node.to_string())
			elif self.node_type == NodeType.concatenate:
				if _latex_format:
					return "{0} \comp {1}".format(self.left_node.to_string(), self.right_node.to_string())
				else:
					return "{0};{1}".format(self.left_node.to_string(), self.right_node.to_string())
			elif self.node_type == NodeType.plus:
				if _latex_format:
					return "({0})^+".format(self.left_node.to_string())
				else:
					return "({0})+".format(self.left_node.to_string())
			elif self.node_type == NodeType.star:
				if _latex_format:
					return "({0})^*".format(self.left_node.to_string())
				else:
					return "({0})*".format(self.left_node.to_string())
			else:
				return "UNKNOWN"
		else:
			# if the path condition doesn't contain a path condition in the first slot (i.e contains a string)
			if self.node_type == NodeType.relation:
				if self.parameter_variable is not None :
					new_string = self.relation_label + "(" + self.parameter_variable + ")"
					return "%s" %new_string
				else :
					return "%s" %self.relation_label
			elif self.node_type == NodeType.reverse:
				if _latex_format:
					return "\overline{{{0}}}".format(self.relation_label)
				else:
					return "¬{0}".format(self.relation_label)
			else:
				return "UNKNOWN"

	def head(self):
		"""Returns the head of a path condition"""
		# ====== debug start ======
		#debugprint(5, "H: " + self.to_string(), self)
		# ====== debug end ======

		if self.node_type == NodeType.relation:
			return [PathCondition(None, None, self.relation_label, NodeType.relation, self.parameter_variable)]
		elif self.node_type == NodeType.reverse:
			if self.left_node.node_type == NodeType.relation:
				return [self.copy()]
			else:
				# push reverse down towards leaves
				self = self.left_node.reverse()
				return self.head()
		elif self.node_type == NodeType.concatenate:
			return self.left_node.head()
		elif self.node_type == NodeType.plus:
			return self.left_node.head()
		elif self.node_type == NodeType.star:
			_head_list = self.left_node.head()
			# insert shorter path conditions first in the list (optimisation)
			_head_list.insert(0, PathCondition(None, None, "\diamond", NodeType.relation,self.parameter_variable))
			return _head_list
		else:
			raise Exception("Unexpected node_type for path condition with first slot containing a path condition")

	def suffix(self, level=0):
		"""Returns the suffix of a path condition"""
		# ====== debug start ======
		#debugprint(5, self.to_string(), self)
		#debugprint(5, "Level: " + str(level) + ", NodeType: " + NodeType.to_string(self.node_type))
		# ====== debug end ======

		# if relation
		if self.node_type == NodeType.relation:
			# if the path condition doesn't contain a path condition in the first slot (i.e contains a string)
			# only return \diamond if at lowest level (i.e. end of suffix) or at next level if last path condition is reversed
			if level == 0:
				return [PathCondition(None, None, "\diamond", NodeType.relation, self.parameter_variable)]
			else:
				return [PathCondition(None, None, "BLANK", NodeType.relation, self.parameter_variable)]
		elif self.node_type == NodeType.reverse:
			if self.left_node.node_type == NodeType.relation:
				if level == 0:
					return [PathCondition(None, None, "\diamond", NodeType.relation,  self.parameter_variable)]
				else:
					return [PathCondition(None, None, "BLANK", NodeType.relation, self.parameter_variable)]
			else:
				# push reverse down towards leaves
				self = self.left_node.reverse()
				#level += 1
				return self.suffix(level)
				#level -= 1
		elif self.node_type == NodeType.concatenate:
			level += 1
			_suffix_list = self.left_node.suffix(level)
			level -= 1
			for place in range(len(_suffix_list)):
				# if BLANK (BLANK put in place of \diamond as not lowest level)
				if type(_suffix_list[place]) == PathCondition and _suffix_list[place].relation_label == "BLANK":
					_suffix_list[place].left_node = self.right_node.left_node
					_suffix_list[place].right_node = self.right_node.right_node
					_suffix_list[place].relation_label = self.right_node.relation_label
					_suffix_list[place].node_type = self.right_node.node_type
					_suffix_list[place].parameter_variable = self.right_node.parameter_variable
				else:
					_suffix_list[place] = PathCondition(_suffix_list[place], self.right_node, None, NodeType.concatenate, self.parameter_variable)
			return _suffix_list
		elif self.node_type == NodeType.plus:
			self.node_type = NodeType.relation
			level += 1
			_suffix_list = self.left_node.suffix(level)
			level -= 1
			self.node_type = NodeType.star
			for place in range(len(_suffix_list)):
				# if BLANK (BLANK put in place of \diamond as not lowest level)
				if type(_suffix_list[place]) == PathCondition and _suffix_list[place].relation_label == "BLANK":
					_suffix_list[place].left_node = self.left_node
					_suffix_list[place].right_node = self.right_node
					_suffix_list[place].relation_label = self.relation_label
					_suffix_list[place].node_type = self.node_type
					_suffix_list[place].parameter_variable = self.parameter_variable
				else:
					_suffix_list[place] = PathCondition(_suffix_list[place], self, None, NodeType.concatenate, self.parameter_variable, self.parameter_constant)
			return _suffix_list
		elif self.node_type == NodeType.star:
			self.node_type = NodeType.relation
			level += 1
			_suffix_list = self.left_node.suffix(level)
			level -= 1
			self.node_type = NodeType.star
			for place in range(len(_suffix_list)):
				# if BLANK (BLANK put in place of \diamond as not lowest level)
				if type(_suffix_list[place]) == PathCondition and _suffix_list[place].relation_label == "BLANK":
					_suffix_list[place].left_node = self.left_node
					_suffix_list[place].right_node = self.right_node
					_suffix_list[place].relation_label = self.relation_label
					_suffix_list[place].node_type = self.node_type
					_suffix_list[place].parameter_variable = self.parameter_variable
				else:
					_suffix_list[place] = PathCondition(_suffix_list[place], self, None, NodeType.concatenate, self.parameter_variable)
			# only append \diamond if at lowest level (i.e. end of suffix) or at next level if last path condition is reversed
			if level == 0:
				# insert shorter path conditions first in the list (optimisation)
				_suffix_list.insert(0, PathCondition(None, None, "\diamond", NodeType.relation, None, self.parameter_variable))
			elif level == 1 and self.node_type == NodeType.reverse:
				# insert shorter path conditions first in the list (optimisation)
				_suffix_list.insert(0, PathCondition(None, None, "\diamond", NodeType.relation, self.parameter_variable))
			else:
				# insert shorter path conditions first in the list (optimisation)
				_suffix_list.insert(0, PathCondition(None, None, "BLANK", NodeType.relation, self.parameter_variable))
			return _suffix_list
		else:
			raise Exception("Unexpected node_type for path condition with first slot containing a path condition")

	def reverse(self):
		# ====== debug start ======
		#debugprint(5, "R: " + self.to_string(), self)
		# ====== debug end ======

		"""Reverses the path condition (modifying the path condition object)"""
		if self.node_type == NodeType.relation:
			# manually wrap the relation in a reverse PathCondition
			self.left_node = PathCondition(None, None, self.relation_label, NodeType.relation,self.parameter_variable)
			self.right_node = None
			self.relation_label = None
			self.node_type = NodeType.reverse
		elif self.node_type == NodeType.reverse:
			if self.left_node.node_type == NodeType.relation:
				# Don't reverse the relation inside this PathCondtion's reverse, instead remove the reverse at this level
				self.relation_label = self.left_node.relation_label
				self.left_node = None
				self.right_node = None
				self.node_type = NodeType.relation
			else:
				# Trigger Reverse on the reversed path condition
				self.left_node.reverse()
		elif self.node_type == NodeType.concatenate:
			# manually reverse the concatenation
			temp = PathCondition(self.left_node, None, None, NodeType.reverse,self.parameter_variable)
			self.left_node = PathCondition(self.right_node, None, None, NodeType.reverse, self.parameter_variable)
			self.right_node = temp
			# reverse the concatenated path conditions should they be concatenated
			if type(self.left_node) == type(self):
				if self.left_node.node_type == NodeType.concatenate:
					self.left_node.reverse()
			if type(self.right_node) == type(self):
				if self.right_node.node_type == NodeType.concatenate:
					self.right_node.reverse()
		elif self.node_type == NodeType.plus:
			self.left_node = PathCondition(self.left_node, None, None, NodeType.reverse,self.parameter_variable)
		elif self.node_type == NodeType.star:
			self.left_node = PathCondition(self.left_node, None, None, NodeType.reverse,self.parameter_variable)
		else:
			raise Exception("Unexpected node_type for path condition with first slot containing a path condition")

	def re_write(self):
		"""Prints the re-written path condition - for testing purposes"""
		_starting_string = self.to_string()
		_head_list = self.head()
		for _place in range(len(_head_list)):
			print("\t\\[{0}: \\textsf{{Head}}({1}) = {2}\\]".format(_place, _starting_string, _head_list[_place].to_string()))
			print("\t\\[{0}: \\textsf{{Head}}({1}) = {2}\\]".format(_place, _starting_string, _head_list[_place].to_string()))
		_suffix_list = self.suffix()
		for _place in range(len(_suffix_list)):
			print("\t\\[{0}: \\textsf{{Suffix}}({1}) = {2}\\]".format(_place, self.to_string(), _suffix_list[_place].to_string()))
		print("")

	def copy(self):
		"""Creates a copy of a path condition in its own memory"""
		# ====== debug start ======
		#debugprint(0, "COPY: " + self.to_string(), self)
		# ====== debug end ======
		if self.node_type == NodeType.relation:
			return PathCondition(None, None, self.relation_label, NodeType.relation, self.parameter_variable)
		elif self.node_type == NodeType.concatenate:
			return PathCondition(self.left_node.copy(), self.right_node.copy(), None, NodeType.concatenate, self.parameter_variable)
		else:
			return PathCondition(self.left_node.copy(), None, None, self.node_type, self.parameter_variable)

	def evaluateParameters(self, m, para, p, new_map):
		if len(para) != len(p):
			return False
		index = 0
		parameterMatchFailed = False
		for par in para:
			if ((par[0].isupper())):
				for k, v in m.items():
					par_exist = False
					if (k == par):
						par_exist = True
						if (v != p[index]):
							parameterMatchFailed = True
							break
				if par_exist == False:
					new_map.update({par: p[index]})
					#m.update({par: p[index]})
			else:
				if (par != p[index]):
					parameterMatchFailed = True
			if parameterMatchFailed == True:
				break
			index = index + 1
		return parameterMatchFailed


	def determine_suffix(self, phi_r, currNode, nextNode, parameterMap, debugprintlevel=0):
		"""Determine Suffix algorithm"""
		# ====== debug start ======
		debugprint(4 + debugprintlevel, "DS Path: " + self.to_string() + ", looking at: " + phi_r.to_string(), self)
		# ====== debug end ======
		#1: Initialize empty list of node, suffix tuples Θ
		Theta = []
		#2: Initialize empty list of head elements Π h
		Pi_h = []
		#3: Initialize empty list of suffixes Π s
		Pi_s = []
		#4: Π h = head(φ)
		Pi_h = self.head()
		#5: Π s = suffix(φ)
		Pi_s = self.suffix()
		#6: Initialize integer i to 0
		i = 0
		#7: while i < number of items in Π h do
		while i < len(Pi_h):
			# ====== debug start ======
			debugprint(4 + debugprintlevel, "Pi_h[" + str(i) + "]: " + Pi_h[i].to_string(), Pi_h[i])
			debugprint(4 + debugprintlevel, "Pi_s[" + str(i) + "]: " + Pi_s[i].to_string(), Pi_s[i])
			# ====== debug end ======
			#8: if Π h [i] = φ r then

			if Pi_h[i].node_type == NodeType.relation:
				if Pi_h[i].left_node == phi_r.left_node and Pi_h[i].right_node == phi_r.right_node and Pi_h[i].relation_label == phi_r.relation_label and Pi_h[i].node_type == phi_r.node_type:
					#9: Θ = Θ join (w, Π s [i])
					parameter_list =[]
					if Pi_h[i].parameter_variable != "*":
						if Pi_h[i].parameter_variable is not None:
							parameter_list = Pi_h[i].parameter_variable.split(",")
						parameter_value_list = []
						if phi_r.parameter_variable is not None:
							parameter_value_list = phi_r.parameter_variable.split(",")
						modifiedMap = copy.deepcopy(parameterMap)
						parameterMatchFailed = self.evaluateParameters(parameterMap, parameter_list, parameter_value_list,
																 modifiedMap)
						if parameterMatchFailed == False :
							Theta.append(VisitedNode(nextNode, Pi_s[i], modifiedMap))
					else :
						Theta.append(VisitedNode(nextNode, Pi_s[i], parameterMap))
				#10: else if Π h [i] = \diamond then
				elif Pi_h[i].relation_label == "\diamond" and Pi_h[i].left_node == None and Pi_h[i].right_node == None:
					#11: if Π s [i] != \diamond then
					if Pi_s[i].relation_label != "\diamond":
						#12: Θ = Θ t DetermineSuffixes(Π s [i], φ r , h, w)
						Theta.extend(Pi_s[i].determine_suffix(phi_r, currNode, nextNode, parameterMap, debugprintlevel + 1))
					#13: else
					else:
						#14: Θ = Θ join [(h, \diamond)]
						Theta.append(VisitedNode(currNode, PathCondition(None, None, "\diamond", NodeType.relation, self.parameter_variable), parameterMap))
					#15: end if
				#16: end if
			elif Pi_h[i].node_type == NodeType.reverse and phi_r.node_type == NodeType.reverse:
				if Pi_h[i].left_node.node_type == NodeType.relation and phi_r.left_node.node_type == NodeType.relation:
					if Pi_h[i].left_node.left_node == phi_r.left_node.left_node and Pi_h[i].left_node.right_node == phi_r.left_node.right_node and Pi_h[i].left_node.relation_label == phi_r.left_node.relation_label and Pi_h[i].right_node == phi_r.right_node and Pi_h[i].relation_label == phi_r.relation_label:
						#9: Θ = Θ join (w, Π s [i])
						Theta.append(VisitedNode(nextNode, Pi_s[i],parameterMap))
			#17: i + +
			i += 1
		#18: end while
		#19: return Θ
		# ====== debug start ======
		for theta in Theta:
			if type(theta.pathCondition) == PathCondition:
				debugprint(4 + debugprintlevel, "DS Returning: " + theta.pathCondition.to_string() + " at Node: " + theta.node, theta.pathCondition)
			else:
				debugprint(4 + debugprintlevel, "DS Returning: " + theta.pathCondition + " at Node: " + theta.node, theta.pathCondition)
		# ====== debug end ======
		return Theta
